#include <Eigen/Core>
#include <ros/ros.h>
#include <ur5/utilities.h>
#include <ur5/ur5_joints.h>
#include <me530646_lab1/lab1.h>
#include <me530646_lab2/lab2.h>
#include <me530646_lab3/lab3.h>

#define PI 3.1415926

//An array of the six alpha D-H parameters for the UR5
double UR5::alpha[] = {PI/2,0,0,PI/2,-PI/2,0};
//An array of the six a D-H parameters for the UR5
double UR5::a[] = {0,-0.425,-0.392,0,0,0};
//An array of the six d D-H parameters for the UR5
double UR5::d[] = {0.0892,0,0,0.1093,0.09475,0.0825};


/**
 * Creates an instance of the class to allow plotting 
 * of the UR5 in rviz.
 */
UR5::UR5(ros::NodeHandle &n):rvizPlotter(n){
 	fwd_pb = n.advertise<ur5::ur5_joints>("forwardKinematics", 10);
	link_names.push_back("shoulder");
	link_names.push_back("upper_arm");
	link_names.push_back("forearm");
	link_names.push_back("wrist_1");
	link_names.push_back("wrist_2");
	link_names.push_back("wrist_3");
	link_names.push_back("end-effector");
}

/**
 * Returns the 4x4 transformation given the a 4x1 vector of
 * Denavit-Hartenberg parameters.
 */
Eigen::Matrix4f UR5::dhf(double alpha, double a, double d, double theta){
	Eigen::Matrix4f homotransf(4,4);
	homotransf<<cos(theta), -sin(theta)*cos(alpha), sin(theta)*sin(alpha), a*cos(theta),
		    sin(theta), cos(theta)*cos(alpha), -cos(theta)*sin(alpha), a*sin(theta),   
		    0, sin(alpha), cos(alpha), d,   
		    0, 0, 0, 1;
	return homotransf;   

	//Done
}

/**
 * Returns the 4x4 transformation from the base of the
 * UR5 to its gripper given a 6x1 vector of the joint
 * angles.
 */
Eigen::Matrix4f UR5::fwd(double q[]){
	Eigen::Matrix4f homotransf1(4,4);     
	Eigen::Matrix4f homotransf2(4,4);
	Eigen::Matrix4f homotransf3(4,4);
	Eigen::Matrix4f homotransf4(4,4);
	Eigen::Matrix4f homotransf5(4,4);
	Eigen::Matrix4f homotransf6(4,4);
	Eigen::Matrix4f homotransfbg(4,4);
	homotransf1=dhf(alpha[0], a[0], d[0], q[0]);
	homotransf2=dhf(alpha[1], a[1], d[1], q[1]);
	homotransf3=dhf(alpha[2], a[2], d[2], q[2]);
	homotransf4=dhf(alpha[3], a[3], d[3], q[3]);
	homotransf5=dhf(alpha[4], a[4], d[4], q[4]);
	homotransf6=dhf(alpha[5], a[5], d[5], q[5]);
	homotransfbg=homotransf1*homotransf2*homotransf3*homotransf4*homotransf5*homotransf6;   
	return homotransfbg;
	//Done     
}

/**
 * Plots the robot in rviz by defining its joint angles.
 */
 void UR5::plotfwd(double q[]){
	ur5::ur5_joints j;

	j.q.push_back(q[0]-PI);
	for(int i = 1; i < 6; i++)
	{
	  j.q.push_back(q[i]);
	}
	while(fwd_pb.getNumSubscribers() < 1){}
	fwd_pb.publish(j);
}

/**
 * Plots the coordinate frames (according to D-H convention) for the
 * UR5 given a set of joint angles. Each frame is plotted relative to
 * fixed frame base_link.
 */	
 void UR5::plotframes(RvizPlotter &plotter, double q[]){
	Eigen::Matrix4f homotransf1(4,4);     
	Eigen::Matrix4f homotransf2(4,4);
	Eigen::Matrix4f homotransf3(4,4);
	Eigen::Matrix4f homotransf4(4,4);
	Eigen::Matrix4f homotransf5(4,4);
	Eigen::Matrix4f homotransf6(4,4);
	Eigen::Matrix4f homotransfb1(4,4);
	Eigen::Matrix4f homotransfb2(4,4);
	Eigen::Matrix4f homotransfb3(4,4);
	Eigen::Matrix4f homotransfb4(4,4);
	Eigen::Matrix4f homotransfb5(4,4);
	Eigen::Matrix4f homotransfbg(4,4);
	homotransf1=dhf(alpha[0], a[0], d[0], q[0]);
	homotransf2=dhf(alpha[1], a[1], d[1], q[1]);
	homotransf3=dhf(alpha[2], a[2], d[2], q[2]);
	homotransf4=dhf(alpha[3], a[3], d[3], q[3]);
	homotransf5=dhf(alpha[4], a[4], d[4], q[4]);
	homotransf6=dhf(alpha[5], a[5], d[5], q[5]);
	homotransfb1=homotransf1;
	homotransfb2=homotransf1*homotransf2;
	homotransfb3=homotransf1*homotransf2*homotransf3;
	homotransfb4=homotransf1*homotransf2*homotransf3*homotransf4;
	homotransfb5=homotransf1*homotransf2*homotransf3*homotransf4*homotransf5;
	homotransfbg=homotransf1*homotransf2*homotransf3*homotransf4*homotransf5*homotransf6;
	plotter.plotf(homotransfb1,"base_link","frame1");     
	plotter.plotf(homotransfb2,"base_link","frame2");         
	plotter.plotf(homotransfb3,"base_link","frame3");
	plotter.plotf(homotransfb4,"base_link","frame4");
	plotter.plotf(homotransfb5,"base_link","frame5");
	plotter.plotf(homotransfbg,"base_link","frame6");
 	//Done          
 	//For plotting use: this->rvizPlotter.plotf(...)
}


Eigen::MatrixXf UR5::jointsposition(double q[]){  //to get the joints positons assuming that each joint  
        Eigen::MatrixXf jposition(6,3);           //lies approximately at the origin of the frame   
	Eigen::Matrix4f homotransf1(4,4);     
	Eigen::Matrix4f homotransf2(4,4);
	Eigen::Matrix4f homotransf3(4,4);
	Eigen::Matrix4f homotransf4(4,4);
	Eigen::Matrix4f homotransf5(4,4);
	Eigen::Matrix4f homotransf6(4,4);
	Eigen::Matrix4f homotransfb1(4,4);
	Eigen::Matrix4f homotransfb2(4,4);
	Eigen::Matrix4f homotransfb3(4,4);
	Eigen::Matrix4f homotransfb4(4,4);
	Eigen::Matrix4f homotransfb5(4,4);
	Eigen::Matrix4f homotransfbg(4,4);
	homotransf1=dhf(alpha[0], a[0], d[0], q[0]);
	homotransf2=dhf(alpha[1], a[1], d[1], q[1]);
	homotransf3=dhf(alpha[2], a[2], d[2], q[2]);
	homotransf4=dhf(alpha[3], a[3], d[3], q[3]);
	homotransf5=dhf(alpha[4], a[4], d[4], q[4]);
	homotransf6=dhf(alpha[5], a[5], d[5], q[5]);
	homotransfb1=homotransf1;
	homotransfb2=homotransf1*homotransf2;
	homotransfb3=homotransf1*homotransf2*homotransf3;
	homotransfb4=homotransf1*homotransf2*homotransf3*homotransf4;
	homotransfb5=homotransf1*homotransf2*homotransf3*homotransf4*homotransf5;
	homotransfbg=homotransf1*homotransf2*homotransf3*homotransf4*homotransf5*homotransf6;
	jposition<<homotransfb1(0,3),homotransfb1(1,3),homotransfb1(2,3),
		   homotransfb2(0,3),homotransfb2(1,3),homotransfb2(2,3),
		   homotransfb3(0,3),homotransfb3(1,3),homotransfb3(2,3),
		   homotransfb4(0,3),homotransfb4(1,3),homotransfb4(2,3),
		   homotransfb5(0,3),homotransfb5(1,3),homotransfb5(2,3),
		   homotransfbg(0,3),homotransfbg(1,3),homotransfbg(2,3);
	return jposition;     

}


